import re

from django.template import Library, TemplateSyntaxError
from django.template.base import FILTER_SEPARATOR, token_kwargs
from django.template.loader_tags import construct_relative_path

from compose_tags.node import ComposeNode, DefineForNode, DefineNode

register = Library()


@register.tag("compose")
def do_compose(parser, token):
    bits = token.split_contents()
    if len(bits) < 2:
        raise TemplateSyntaxError(
            "%r tag takes at least one argument: the name of the template to "
            "be included." % bits[0]
        )
    template_bit = construct_relative_path(parser.origin.template_name, bits[1])
    remaining_bits = bits[2:]

    takes_context = "takes_context" in remaining_bits
    extra_context = token_kwargs(remaining_bits, parser)
    if "children" in extra_context:
        raise TemplateSyntaxError(
            "%r must not take children as a keyword argument." % bits[0]
        )
    nodelist = parser.parse((f"endcompose",))
    parser.next_token()

    return ComposeNode(
        parser.compile_filter(template_bit),
        nodelist,
        extra_context,
        takes_context,
    )


@register.tag("define")
def do_define(parser, token):
    bits = token.split_contents()
    if len(bits) != 2:
        raise TemplateSyntaxError(
            "define tag takes exactly one argument: the name of the template variable that should store the result. Eg: {% define myvar %}value{% enddefine %}"
        )
    target_var = bits[1]

    nodelist = parser.parse((f"enddefine",))
    parser.next_token()
    return DefineNode(target_var, nodelist)


@register.tag("definelist")
def do_define_list(parser, token):
    bits = token.split_contents()
    if len(bits) < 2:
        raise TemplateSyntaxError(
            "%r tag takes at least one argument: the name of the template variable that should store the result. Eg: {% definelist myvar for x in list %}item is {x}{% enddefinelist %}"
            % bits[0]
        )
    target_var = bits[1]

    if len(bits) == 2 or bits[2] != "for":
        raise TemplateSyntaxError(
            "definelist statements should be formated as {% definelist myvar for x in list %}"
        )

    # Most of this implementation come from django.template.defaulttags.do_for
    for_bits = bits[2:]
    is_reversed = for_bits[-1] == "reversed"
    in_index = -3 if is_reversed else -2
    if for_bits[in_index] != "in":
        raise TemplateSyntaxError(
            "'define myvar for' statements should use the format"
            " 'define myvar for x in y': %s" % token.contents
        )

    invalid_chars = frozenset((" ", '"', "'", FILTER_SEPARATOR))
    loopvars = re.split(r" *, *", " ".join(for_bits[1:in_index]))
    for var in loopvars:
        if not var or not invalid_chars.isdisjoint(var):
            raise TemplateSyntaxError(
                "'define myvar for' received an invalid argument:"
                " %s" % token.contents
            )

    sequence = parser.compile_filter(for_bits[in_index + 1])
    nodelist_loop = parser.parse(
        (
            "empty",
            "enddefinelist",
        )
    )
    token = parser.next_token()
    if token.contents == "empty":
        nodelist_empty = parser.parse(("enddefinelist",))
        parser.delete_first_token()
    else:
        nodelist_empty = None
    return DefineForNode(
        target_var, loopvars, sequence, is_reversed, nodelist_loop, nodelist_empty
    )



#!/usr/bin/env python3
import os
import re
import shutil
import sys
from io import open

from setuptools import find_packages, setup


def read(f):
    return open(f, "r", encoding="utf-8").read()


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, "__init__.py")).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


version = get_version("compose_tags")


if sys.argv[-1] == "publish":
    if os.system("pip freeze | grep twine"):
        print("twine not installed.\nUse `pip install twine`.\nExiting.")
        sys.exit()
    os.system("python setup.py sdist bdist_wheel")
    if os.system("twine check dist/*"):
        print("twine check failed. Packages might be outdated.")
        print("Try using `pip install -U twine wheel`.\nExiting.")
        sys.exit()
    os.system("twine upload dist/*")
    print("You probably want to also tag the version now:")
    print("  git tag -a %s -m 'version %s'" % (version, version))
    print("  git push --tags")
    shutil.rmtree("dist")
    shutil.rmtree("build")
    shutil.rmtree("django_compose_tags.egg-info")
    sys.exit()


setup(
    name="django-compose-tags",
    version=version,
    url="https://github.com/HelloWatt/django-compose-tags",
    license="BSD-3-Clause",
    description="Composition for django templates.",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    author="Jérôme Bon",
    author_email="jerome@hellowatt.fr",
    packages=find_packages(exclude=["tests*"]),
    include_package_data=True,
    install_requires=["Django>=3.0"],
    python_requires=">=3.6",
    zip_safe=False,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
        "Framework :: Django :: 3.1",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.0",
        "Framework :: Django :: 4.1",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
    project_urls={
        "Source": "https://github.com/HelloWatt/django-compose-tags",
        "Changelog": "https://github.com/HelloWatt/django-compose-tags/blob/main/CHANGELOG.md",
    },
)



from django.template import Node, NodeList
from django.template.library import InclusionNode
from django.template.loader_tags import construct_relative_path
from django.utils.safestring import mark_safe

COMPOSE_CONTEXT_KEY = "_django_compose_context_key"


class ComposeNode(Node):
    def __init__(
        self,
        template,
        nodelist,
        extra_context,
        takes_context,
    ):
        super().__init__()
        self.nodelist = nodelist
        self.template = template
        self.extra_context = extra_context or {}
        self.takes_context = takes_context

    def render(self, context):
        template = self.get_template(context)
        render_context = self.get_render_context(context)
        if self.takes_context:
            with context.push(**render_context):
                return template.render(context)
        return template.render(context.new(render_context))

    def get_template(self, context):
        """Very similar implementation to django.template.loader_tags.IncludeNode"""
        template = self.template.resolve(context)
        # Does this quack like a Template?
        if not callable(getattr(template, "render", None)):
            # If not, try the cache and select_template().
            template_name = template or ()
            if isinstance(template_name, str):
                template_name = (
                    construct_relative_path(
                        self.origin.template_name,
                        template_name,
                    ),
                )
            else:
                template_name = tuple(template_name)
            cache = context.render_context.dicts[0].setdefault(self, {})
            template = cache.get(template_name)
            if template is None:
                template = context.template.engine.select_template(template_name)
                cache[template_name] = template
        # Use the base.Template of a backends.django.Template.
        elif hasattr(template, "template"):
            template = template.template
        return template

    def get_render_context(self, context):
        values = {
            name: var.resolve(context) for name, var in self.extra_context.items()
        }
        values["children"] = self.nodelist.render(context)
        # Copy across the CSRF token, if present, because we need instructions for using CSRF
        # protection to be as simple as possible.
        if not self.takes_context:
            csrf_token = context.get("csrf_token")
            if csrf_token is not None:
                values["csrf_token"] = csrf_token
        return values


class CompositionNode(InclusionNode):
    def __init__(self, func, takes_context, args, kwargs, filename, nodelist):
        super().__init__(func, takes_context, args, kwargs, filename)
        self.nodelist = nodelist

    def get_resolved_arguments(self, context):
        resolved_args, resolved_kwargs = super().get_resolved_arguments(context)
        children = self.nodelist.render(context)
        resolved_args = [children] + resolved_args
        return resolved_args, resolved_kwargs


class DefineNode(Node):
    def __init__(self, target_var, nodelist):
        self.target_var = target_var
        self.nodelist = nodelist

    def render(self, context):
        context[self.target_var] = self.nodelist.render(context)
        return ""


class DefineForNode(Node):
    # copypasta of django.template.defaulttags.ForNode with small modifications for target_var
    child_nodelists = ("nodelist_loop", "nodelist_empty")

    def __init__(
        self,
        target_var,
        loopvars,
        sequence,
        is_reversed,
        nodelist_loop,
        nodelist_empty=None,
    ):
        self.target_var = target_var  # divergence from ForNode
        self.loopvars, self.sequence = loopvars, sequence
        self.is_reversed = is_reversed
        self.nodelist_loop = nodelist_loop
        if nodelist_empty is None:
            self.nodelist_empty = NodeList()
        else:
            self.nodelist_empty = nodelist_empty

    def __repr__(self):
        reversed_text = " reversed" if self.is_reversed else ""
        return "<%s: %s, for %s in %s, tail_len: %d%s>" % (
            self.__class__.__name__,
            self.target_var,  # divergence from ForNode
            ", ".join(self.loopvars),
            self.sequence,
            len(self.nodelist_loop),
            reversed_text,
        )

    def render(self, context):
        list_value = []  # divergence from ForNode

        if "forloop" in context:
            parentloop = context["forloop"]
        else:
            parentloop = {}

        with context.push():
            values = self.sequence.resolve(context, ignore_failures=True)
            if values is None:
                values = []
            if not hasattr(values, "__len__"):
                values = list(values)
            len_values = len(values)
            if len_values < 1:
                list_value = [
                    self.nodelist_empty.render(context)
                ]  # divergence from ForNode
            else:  # divergence from ForNode
                if self.is_reversed:
                    values = reversed(values)
                num_loopvars = len(self.loopvars)
                unpack = num_loopvars > 1
                # Create a forloop value in the context.  We'll update counters on each
                # iteration just below.
                loop_dict = context["forloop"] = {"parentloop": parentloop}
                for i, item in enumerate(values):
                    nodelist = []  # divergence from ForNode
                    # Shortcuts for current loop iteration number.
                    loop_dict["counter0"] = i
                    loop_dict["counter"] = i + 1
                    # Reverse counter iteration numbers.
                    loop_dict["revcounter"] = len_values - i
                    loop_dict["revcounter0"] = len_values - i - 1
                    # Boolean values designating first and last times through loop.
                    loop_dict["first"] = i == 0
                    loop_dict["last"] = i == len_values - 1

                    pop_context = False
                    if unpack:
                        # If there are multiple loop variables, unpack the item into
                        # them.
                        try:
                            len_item = len(item)
                        except TypeError:  # not an iterable
                            len_item = 1
                        # Check loop variable count before unpacking
                        if num_loopvars != len_item:
                            raise ValueError(
                                "Need {} values to unpack in for loop; got {}. ".format(
                                    num_loopvars, len_item
                                ),
                            )
                        unpacked_vars = dict(zip(self.loopvars, item))
                        pop_context = True
                        context.update(unpacked_vars)
                    else:
                        context[self.loopvars[0]] = item

                    for node in self.nodelist_loop:
                        nodelist.append(node.render_annotated(context))

                    if pop_context:
                        # Pop the loop variables pushed on to the context to avoid
                        # the context ending up in an inconsistent state when other
                        # tags (e.g., include and with) push data to context.
                        context.pop()
                    list_value.append(
                        mark_safe("".join(nodelist))
                    )  # divergence from ForNode
        context[self.target_var] = list_value  # divergence from ForNode
        return ""  # divergence from ForNode



import functools
from inspect import getfullargspec, unwrap

import django.template
from django.template import TemplateSyntaxError
from django.template.library import parse_bits

from compose_tags.node import CompositionNode


def default_composition(children, **kwargs):
    kwargs["children"] = children
    return kwargs


def parse_bits_with_children(
    parser,
    bits,
    params,
    varargs,
    varkw,
    defaults,
    kwonly,
    kwonly_defaults,
    takes_context,
    name,
):
    """
    Do the same as parse_bits, except that context is the second argument, not the first.
    The first one being children.
    """
    if not params or not params[0] == "children":
        raise TemplateSyntaxError(
            "'%s' must have a first argument of 'children'" % name
        )
    params = params[1:]
    if takes_context:
        if params and params[0] == "context":
            params = params[1:]
        else:
            raise TemplateSyntaxError(
                "'%s' is decorated with takes_context=True so it must "
                "have a second argument of 'context'" % name
            )
    return parse_bits(
        parser,
        bits,
        params,
        varargs,
        varkw,
        defaults,
        kwonly,
        kwonly_defaults,
        False,
        name,
    )


def composition_tag(
    filename,
    takes_context=False,
):
    """
    Register a callable as a composition tag:

    @register.tag
    @composition_tag('card.html', takes_context=True)
    def card(children, context, footer):
        return {
            "children": children,
            "footer": footer or default_footer,
        }
    """

    def dec(func_or_parser, token=None):
        func = default_composition if token else func_or_parser
        (
            params,
            varargs,
            varkw,
            defaults,
            kwonly,
            kwonly_defaults,
            _,
        ) = getfullargspec(unwrap(func))

        @functools.wraps(func)
        def compile_func(parser, token):
            name, *bits = token.split_contents()
            args, kwargs = parse_bits_with_children(
                parser,
                bits,
                params,
                varargs,
                varkw,
                defaults,
                kwonly,
                kwonly_defaults,
                takes_context,
                name,
            )
            nodelist = parser.parse((f"end{name}",))
            parser.next_token()

            return CompositionNode(
                func, takes_context, args, kwargs, filename, nodelist
            )

        if token:
            return compile_func(func_or_parser, token)
        return compile_func

    dec.__name__ = ".".join(filename.split("/")[-1].split(".")[:-1])

    return dec



import django

__title__ = "Django Compose Tag"
__version__ = "0.0.1"
__author__ = "Jérôme Bon"
__license__ = "BSD 3-Clause"
__copyright__ = "Copyright 2022 Hello Watt"

# Version synonym
VERSION = __version__

from .tag import composition_tag
